;---------------------------------------------------
; This program is distributed under GNU GPL license
;---------------------------------------------------
	processor p16f84
	__config 0x3ff2
	radix	dec

w	equ	0
f	equ	1

c	equ	0
dc	equ	1
z	equ	2

indf	equ	0x0
tmr0	equ	0x1
pcl	equ	0x2
status	equ	0x3
fsr	equ	0x4
porta	equ	0x5
portb	equ	0x6
eedata	equ	0x8
eeadr	equ	0x9
pclatch	equ	0xa
intcon	equ	0xb

optionr	equ	0x81
trisa	equ	0x85
trisb	equ	0x86
eecon1	equ	0x88
eecon2	equ	0x89

;----------------------------------------------
; Variable defs.
;----------------------------------------------
	cblock	0xc

	tmp_w
	tmp_st
        nirq
        nsamples
        sampleix
        comb0
        comb0h
        comb1
        comb1h
        comb2
        comb2h
        comb3
        comb3h
        aout
        aouth
        adata
	adata_h
        nchar
        resto
        cntbit
        obyte
        uart
        uarth
        uartdel
	tmp
        buf

	endc

;------ RESET ----------------------------
;	Configuration:
;	---------------
;	RA3 A/D input
;	RA2 A/D output
;	RB7 UART output
;
	org	0

	movlw	0xfb	;RA2 out
	tris	porta
	goto	ini

;------ INTERRUPT (9615/sec)-----------------

	org	0x4

	movwf	tmp_w		;save CPU status
	swapf	status,w
	movwf	tmp_st

	bcf	intcon,2	;clear interrupt

	;Interrupt latency equalization (to add 1-cycle delay if needed)
	; - can give 3 dB improvement (1/2 bit)
	;  Some experimentation is still needed because 
	;  PIC16F84 interrupt details are not well known
	;We can try:
	; - No equalization
	; - btfsc   tmr0,0
	; - btfss   tmr0,0
	btfsc	tmr0,0
	goto	deleq

deleq	movlw	259-104		;adjust timer for a new
	addwf	tmr0,f		;interrupt

	;------ A/D -----
	rrf	porta,w		;A/D feedback loop
	xorlw	4		;~RA3 -> RA2
	movwf	porta

        ;---- UART ----		;16-bit shift reg.
        bsf     status,c
        rrf     uarth,f
        rrf     uart,f
        btfsc   status,c
        bsf     portb,7
        btfss   status,c
        bcf     portb,7
	movf    uartdel,f         ;dec. uartdel if not 0
        btfss   status,z
        decf    uartdel

        ;---- Comb Filter: integrators ----
	btfsc	porta,2
	goto	isr1
	incfsz	comb0,f
	goto	isr1
	incf	comb0h,f
isr1    movf    comb0,w
        addwf   comb1,f
        btfsc   status,c
        incf    comb1h,f
        movf    comb0h,w
        addwf   comb1h,f

        ;----- DECIMATION: The differentiator code is called
	;----- every 256 interrupts

        incfsz  nirq,f
        goto isrf

	;----- Comb Filter: differentiators -----
        ; aout =comb1-comb2-comb3
	; comb3=comb1-comb2
	; comb2=comb1

        movf    comb1,w
        movwf   aout
        movf    comb1h,w
        movwf   aouth

        movf    comb2,w
        subwf   aout,f
        btfss   status,c
        decf    aouth,f
        movf    comb2h,w
        subwf   aouth,f

        movf    aout,w
        movwf   comb2
        movf    aouth,w
        movwf   comb2h

        movf    comb3,w
        subwf   aout
        btfss   status,c
        decf    aouth,f
        movf    comb3h,w
        subwf   aouth,f

        movf    comb2,w
        movwf   comb3
        movf    comb2h,w
        movwf   comb3h
        movf    comb1,w
        movwf   comb2
        movf    comb1h,w
        movwf   comb2h

        incf    nsamples,f

	;---- End of Interrupt ----
isrf	swapf	tmp_st,w	;restore cpu status
	movwf	status
	swapf	tmp_w,f
	swapf	tmp_w,w
	retfie

;-------------------------------------------
;------ MAIN -------------------------------
;-------------------------------------------

ini	movlw	0x7e	;RB7, RB0 out
	tris	portb

	movlw	0x88	;No Pull-UPs, no prescaler
	option

	clrf	nirq	;Clearing variables
	clrf	nsamples
        clrf    sampleix
        clrf    comb0
        clrf    comb0h
        clrf    comb1
        clrf    comb1h
        clrf    comb2
        clrf    comb2h
        clrf    comb3
        clrf    comb3h
	movlw	0xff
	movwf	uart
	movwf	uarth

	bsf	portb,7	; Default state of UART's line

	movlw	0xa0	; Interrupt on timer overflow
	movwf	intcon

buc	call	getad	; get AD sample
        call	prtn	; And print it in decimal
	goto	buc

;-------- UART EMULATOR (9600 bps) ---------
; Sends the byte: "obyte"
;-------------------------------------------

putbyte movf	nirq,w
	movwf	tmp

pbbuc1	movf	tmp,w	;wait until next interrupt
	subwf	nirq,w
	btfsc   status,z
	goto 	pbbuc1

pbbuc2	movf    uartdel,f
        btfss   status,z
        goto    pbbuc2

        movlw   0xff	; Filling the UART shift reg.
        movwf   uarth
        movf    obyte,w
        movwf   uart
        bcf     status,c
        rlf     uart
        rlf     uarth

        movlw   10	; Wait until 10 interrupts
        movwf   uartdel

	return

;------------------ BIN to ASCII -----------------
; divide by 10: adata(16 bits)/10 -> adata,resto
;-------------------------------------------------

div10	movlw	16
	movwf   cntbit
	clrf	resto
	movlw	10
bdiv1	bcf	status,0
	rlf	adata,f
	rlf	adata_h,f
	rlf	resto,f
	subwf	resto,f
	btfsc	status,c
	incf	adata
	btfss	status,c
	addwf	resto,f
	decfsz  cntbit,f
	goto	bdiv1
	retlw	0

;------------------------------------------------
;-- prtn: prints adata (16 bits) in ascii
;------------------------------------------------

prtn	movlw	buf
	movwf	fsr
	clrf	nchar
bprtn1	call	div10
	incf	nchar
	movf	resto,w
	addlw	48
	movwf	indf
	incf	fsr
	movf	adata,w
	iorwf	adata_h,w
	btfss	status,z
	goto	bprtn1
bprtn2	decf	fsr,f
	movf	indf,w
	movwf	obyte
	call	putbyte
	decfsz	nchar,f
	goto	bprtn2
	movlw	10
	movwf	obyte
	call	putbyte
	retlw	0

;-----------------------------------------------
; A/D read: new sample -> adata
;-----------------------------------------------

getad   movf    nsamples,w
        subwf   sampleix,w
        btfsc   status,z
        goto    getad
        movf    nsamples,w
        movwf   sampleix
        movf    aout,w
        movwf   adata
        movf    aouth,w
        movwf   adata_h
	bsf	portb,0	; Pulse on PB0 for timming debug
	bcf	portb,0
	return
