/*--------------------------------------------------------------------------
laX11.c:
Logic Analizer code.
--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#define  XK_MISCELLANY
#include <X11/keysymdef.h>
//#include <math.h>
#include <string.h>
#include "pic.h"

// Logic Analizer functions: Keep trace of ports A & B

typedef struct {
        long tstamp;
        byte ra;
        byte rb;
} la_event;

la_event la_evbuf[0x10000];
unsigned short la_evp=0;
int la_nev=0;

void la_add_event(core *pcore)
{
        static int nested=0;
        unsigned short ix;

        if (nested) return;
        nested++;
        la_evbuf[la_evp].tstamp=pcore->Clocks;
        la_evbuf[la_evp].ra=(porta_rd(pcore) & pcore->trisa) |
                            (pcore->porta & (~pcore->trisa));
        la_evbuf[la_evp].rb=(portb_rd(pcore) & pcore->trisb) |
                            (pcore->portb & (~pcore->trisb));
        // Compress events (a change is needed to record an event)
        ix=la_evp-1;
        if (la_evbuf[la_evp].ra!=la_evbuf[ix].ra ||
            la_evbuf[la_evp].rb!=la_evbuf[ix].rb    ) {
                la_evp++;
                la_nev++; if (la_nev>0x10000) la_nev=0x10000;
        }
        nested--;
}

// A function to print numbers in engeeniering format from femtos to Gigas.
// Used to print time intervals and frequencies
char *sprteng(float d)
{
        const char sunit[]={0,'f','p','n','u','m',' ','k','M','G',0};
        static char buf[32];
        int i=6;

        while (d<1 && sunit[i-1]) {d*=1000.; i--;}
        while (d>=1000 && sunit[i+1]) {d*=0.001; i++;}
        sprintf(buf,"%3.2f %c",d,sunit[i]);
        return buf;
}

//--------------------------------------------------------------------------
// Logic Analizer interface
//--------------------------------------------------------------------------

char hello[] = {"Logic Analizer"};
extern core f84;

void la_show_events()
{
    Display *mydisplay;
    Window mywindow;
    GC mygc;
    XEvent myevent;
    XSizeHints myhint;
    XWMHints mywmhint;
    int myscreen;
    unsigned long colf,colb;
    XColor p1,p2;
    Colormap cmap;
    int colort[16];

    int i,j,k,x0,x1,ox,oy,x,y,m,h,dh,dh1;
    int mx,ax=0,bx=0;
    float t;
    unsigned short ix,ixb,ixt;
    int ix0,ix0b,ix1,ix1b;
    char text[32];
    int done;

    printf("\r\033[J\033m\033[1mCursors:\033[m center & zoom ");
    printf("\033[1mMouse:\033[m time_marks ");
    printf("\033[1mR\033[meset time_marks \033[1mF\033[mit ");
    printf("\033[1mQ\033[muit ");
    fflush(stdout);

    mydisplay = XOpenDisplay("");
    if(mydisplay==NULL) {fprintf(stderr,"\033[2J\033[HCan't connect to %s\n",
			XDisplayName("")); getchar(); return;}
    myscreen = DefaultScreen(mydisplay);
    colf = WhitePixel(mydisplay,myscreen);
    colb = BlackPixel(mydisplay,myscreen);
    myhint.x=200;myhint.y=0;
    myhint.width=640;myhint.height=480;
    myhint.flags=PPosition|PSize;
    mywmhint.flags=InputHint;
    mywmhint.input=True;

    mywindow=XCreateSimpleWindow(mydisplay,
	DefaultRootWindow(mydisplay),
	myhint.x,myhint.y,myhint.width,myhint.height,
	5,colf,colb);
    XSetStandardProperties(mydisplay,mywindow,hello,hello,
    	None,NULL,0,&myhint);
    XSetWMHints(mydisplay,mywindow,&mywmhint);

    mygc=XCreateGC(mydisplay,mywindow,0,0);
    XSetBackground(mydisplay,mygc,colb);
    XSetForeground(mydisplay,mygc,colf);

    XSelectInput(mydisplay,mywindow,
	ButtonPressMask|KeyPressMask|ExposureMask);

    XMapRaised(mydisplay,mywindow);
    cmap=DefaultColormap(mydisplay,myscreen);

    ixb=la_evp-1;
    x1=la_evbuf[ixb].tstamp;
    ixb-=la_nev-2;
    x0=la_evbuf[ixb].tstamp;

    for(done=0;done<2;) {
        // Show the waveforms for the current time interval (x0:x1)
        // Alloc colors for traces
        XClearWindow(mydisplay,mywindow);
        XAllocNamedColor(mydisplay,cmap,"gray"   ,&p1,&p2); colort[0]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"red"    ,&p1,&p2); colort[1]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"orange" ,&p1,&p2); colort[2]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"yellow" ,&p1,&p2); colort[3]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"green"  ,&p1,&p2); colort[4]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"blue"   ,&p1,&p2); colort[5]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"cyan"   ,&p1,&p2); colort[6]=p1.pixel;
        XAllocNamedColor(mydisplay,cmap,"magenta",&p1,&p2); colort[7]=p1.pixel;

        // Find the border events
        ix0=ix1=0; ix0b=ix1b=la_nev-1;
        for (i=0;i<16;i++) {
                ix=(ix0+ix0b)>>1; ixt=ix+ixb;
                if(la_evbuf[ixt].tstamp>x0) ix0b=ix;
                else ix0=ix;
                ix=(ix1+ix1b)>>1; ixt=ix+ixb;
                if(la_evbuf[ixt].tstamp>x1) ix1b=ix;
                else ix1=ix;
        }

        // Port A Waveforms

        h=2; dh=myhint.height/16; dh1=dh*3/4;
        for (j=0;j<5;j++) {
            XSetForeground(mydisplay,mygc,colort[j]);
            m=(1<<j);
            h+=dh;
            sprintf(text,"RA%d",j);
            XDrawImageString(mydisplay,mywindow,mygc,2,h-4,text,3);

            for (ix=ix0+1;ix<=ix1;ix++) {
                ixt=ix+ixb-1;
                ox=32+(la_evbuf[ixt].tstamp-x0)*(myhint.width-32)/(x1-x0);
                oy=h-((la_evbuf[ixt].ra & m)?dh1:0);
                ixt++;
                x=32+(la_evbuf[ixt].tstamp-x0)*(myhint.width-32)/(x1-x0);
                y=h-((la_evbuf[ixt].ra & m)?dh1:0);
                XDrawLine(mydisplay,mywindow,mygc,ox,oy,x,oy);
                if (oy!=y)XDrawLine(mydisplay,mywindow,mygc,x,oy,x,y);
            }
            if (ix0==ix1) {
                ixt=ix+ixb-1;
                y=h-((la_evbuf[ixt].ra & m)?dh1:0);
                x=32;
            }
            XDrawLine(mydisplay,mywindow,mygc,x,y,myhint.width,y);
        }

        // Port B Waveforms

        h+=2*dh;
        for (j=0;j<8;j++) {
            XSetForeground(mydisplay,mygc,colort[j]);
            m=(1<<j);
            h+=dh;
            sprintf(text,"RB%d",j);
            XDrawImageString(mydisplay,mywindow,mygc,2,h-4,text,3);

            for (ix=ix0+1;ix<=ix1;ix++) {
                ixt=ix+ixb-1;
                ox=32+(la_evbuf[ixt].tstamp-x0)*(myhint.width-32)/(x1-x0);
                oy=h-((la_evbuf[ixt].rb & m)?dh1:0);
                ixt++;
                x=32+(la_evbuf[ixt].tstamp-x0)*(myhint.width-32)/(x1-x0);
                y=h-((la_evbuf[ixt].rb & m)?dh1:0);
                XDrawLine(mydisplay,mywindow,mygc,ox,oy,x,oy);
                if (oy!=y)XDrawLine(mydisplay,mywindow,mygc,x,oy,x,y);
            }
            if (ix0==ix1) {
                ixt=ix+ixb-1;
                y=h-((la_evbuf[ixt].rb & m)?dh1:0);
                x=32;
            }
            XDrawLine(mydisplay,mywindow,mygc,x,y,myhint.width,y);
        }

        // Time limits and time marks

        h+=dh;
        XSetForeground(mydisplay,mygc,colf);
        sprintf(text,"%d",x0);
        XDrawImageString(mydisplay,mywindow,mygc,32,h-4,text,strlen(text));
        sprintf(text,"%d",x1);
        XDrawImageString(mydisplay,mywindow,mygc,
                myhint.width-2-7*strlen(text),h-4,text,strlen(text));

        sprintf(text,"L :%d",ax);
        XDrawImageString(mydisplay,mywindow,mygc,
                myhint.width/8,h-4,text,strlen(text));
        sprintf(text,"R :%d",bx);
        XDrawImageString(mydisplay,mywindow,mygc,
                myhint.width*7/8-7*strlen(text),h-4,text,strlen(text));
        if (ax!=bx) {
            t=abs(bx-ax); t/=f84.freq;
            sprintf(text,"dt :%d (%ss)",bx-ax,sprteng(t));
            sprintf(text,"%s (%sHz)",text,sprteng(1./t));
            XDrawImageString(mydisplay,mywindow,mygc,
                (myhint.width-strlen(text)*7)/2,h-4,text,strlen(text));
        }
        if (ax>x0 && ax<x1) {
                x=32+(ax-x0)*(myhint.width-32)/(x1-x0);
                XDrawLine(mydisplay,mywindow,mygc,x,myhint.height-16,x,0);
        }
        if (bx>x0 && bx<x1) {
                x=32+(bx-x0)*(myhint.width-32)/(x1-x0);
                XDrawLine(mydisplay,mywindow,mygc,x,myhint.height-16,x,0);
        }

        // X11 Events processing

        for(done=0;!done;) {
          XNextEvent(mydisplay,&myevent);
	  switch(myevent.type)
	  {

          case ButtonPress:  // Update time marks
             mx=myevent.xbutton.x;
             switch(((XButtonEvent*)&myevent)->button){
             case 1: ax=x0+(mx-32)*(x1-x0)/(myhint.width-32); break;
             case 3: bx=x0+(mx-32)*(x1-x0)/(myhint.width-32); break;
             }
             done=1; break;
          //case Expose:
          //   done=1; break;
          case KeyPress:
            j=XLookupKeysym((XKeyEvent *)&myevent,0);
            i=(x1-x0)/2;
	    switch(j) {
            case 'q':           // Q pressed: Quit
                done=2; break;
            case 'r':           // R pressed: Reset cursors
                ax=bx=0;
                done=1; break;
            case 'f':           // F pressed: Fit visible
                ixb=la_evp-1;
                x1=la_evbuf[ixb].tstamp;
                ixb-=la_nev-2;
                x0=la_evbuf[ixb].tstamp;
                done=1; break;
            case XK_Up:         // Up-> Zoom
                x0+=i/2;
                x1-=i/2;
                done=1; break;
            case XK_Down:       // Down -> Unzoom
                x0-=i;
                x1+=i;
                done=1; break;
            case XK_Right:      // Right
                x0+=i/4;
                x1+=i/4;
                done=1; break;
            case XK_Left:       // Left
                x0-=i/4;
                x1-=i/4;
                done=1; break;
	    }
	    break;
          }
        }
    }

    XFreeGC(mydisplay,mygc);
    XDestroyWindow(mydisplay,mywindow);
    XCloseDisplay(mydisplay);
}
