/*-------------------------------------------------------------------------
softvideo.c:
Emulator for Rickard Gune Video hardware.
- Time is measured directly in CPU clocks. 12 MHz osc. is assumed.
- Vertical synchronism is detected when two sync pulses are enough close
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#define  XK_MISCELLANY
#include <X11/keysymdef.h>
#include "pic.h"

// IO emulation routines
// External interface emulation

// RD routines. Called very often to check timer & interrupt pins

byte porta_rd(core *pcore)
{
        static byte opa=0;
        byte pa;

        pa=0xff;
        //if (pa!=opa) {opa=pa; la_add_event(pcore);}
        return pa;
}

byte joy=0xff;  // Joystick status

byte portb_rd(core *pcore)
{
        static byte opb=0;
        byte pb;

        pb=joy;

        //if (pb!=opb) {opb=pb; la_add_event(pcore);}
        return pb;
}

// WR routines. Called only on program writes to ports

const char title[] = {"Softvideo Screen"};
extern core f84;

Display *mydisplay;
Window mywindow;
GC mygc;
XEvent myevent;
XSizeHints myhint;
XWMHints mywmhint;
int myscreen;
unsigned long colf,colb;

void initX()
{
    mydisplay = XOpenDisplay("");
    if(mydisplay==NULL) {fprintf(stderr,"\033[2J\033[HCan't connect to %s\n",
			XDisplayName("")); getchar(); return;}
    myscreen = DefaultScreen(mydisplay);
    colf = WhitePixel(mydisplay,myscreen);
    colb = BlackPixel(mydisplay,myscreen);
    myhint.x=200;myhint.y=0;
    myhint.width=192*2;myhint.height=312;
    myhint.flags=PPosition|PSize;
    mywmhint.flags=InputHint;
    mywmhint.input=True;

    mywindow=XCreateSimpleWindow(mydisplay,
	DefaultRootWindow(mydisplay),
	myhint.x,myhint.y,myhint.width,myhint.height,
	5,colf,colb);
    XSetStandardProperties(mydisplay,mywindow,title,title,
    	None,NULL,0,&myhint);
    XSetWMHints(mydisplay,mywindow,&mywmhint);

    mygc=XCreateGC(mydisplay,mywindow,0,0);
    XSetBackground(mydisplay,mygc,colb);
    XSetForeground(mydisplay,mygc,colf);

    XSelectInput(mydisplay,mywindow,
	ButtonPressMask|KeyPressMask|KeyReleaseMask|ExposureMask);

    XMapRaised(mydisplay,mywindow);
    //cmap=DefaultColormap(mydisplay,myscreen);

    XClearWindow(mydisplay,mywindow);
    XFlush(mydisplay);
}

// Function that checks the keyboard every 20ms

void joyupdate()
{
        int j;

        if (!XCheckWindowEvent(mydisplay,mywindow,-1,&myevent)) return;

        switch(myevent.type){
        case KeyPress:
            j=XLookupKeysym((XKeyEvent *)&myevent,0);
	    switch(j) {
            case ' ':           joy&=~2; break;
            case XK_Up:         joy&=~8; break;
            case XK_Down:       joy&=~4; break;
            case XK_Right:      joy&=~16; break;
            case XK_Left:       joy&=~32; break;
	    }
	    break;
        case KeyRelease:
            j=XLookupKeysym((XKeyEvent *)&myevent,0);
	    switch(j) {
            case ' ':           joy|=2; break;
            case XK_Up:         joy|=8; break;
            case XK_Down:       joy|=4; break;
            case XK_Right:      joy|=16; break;
            case XK_Left:       joy|=32; break;
	    }
	    break;
        }
}

// RA0 is used for video sync. Adjust the "electron beam" position
// to the left after an horizontal sync pulse, and up after a vertical
// sync pulse.

unsigned int posx=0,posy=0xffff,htime;

void porta_wr(core *pcore)
{
        static byte firstime=1,opa=0,otra=0xff;
        if (pcore->porta!=opa || pcore->trisa!=otra) la_add_event(pcore);
        if (firstime) {firstime=0; initX();}

        if (!(opa&1) && (pcore->porta&1)) { //Positive edge
                if ((pcore->Clocks-htime)>144) { // Hsync
                        // Ends drawing the previous line
                        XDrawLine(mydisplay,mywindow,mygc,(posx<<1)+1,posy,383,posy);
                        posx=0; posy++;
                } else {        // Vsync
                        posy=posx=0;
                        //joyupdate();
                }
                htime=pcore->Clocks;
        }

        opa=pcore->porta; otra=pcore->trisa;
}

extern int debug_status;

// RB0 is the two-level video output. Draw an horizontal segment from
// The last "spot" position

void portb_wr(core *pcore)
{
        static byte opb=0,otrb=0xff;
        unsigned int x;
        if (pcore->portb!=opb || pcore->trisb!=otrb) la_add_event(pcore);

        if(posy<312) {
            x=pcore->Clocks-htime;
            XDrawLine(mydisplay,mywindow,mygc,(posx<<1)+1,posy,x<<1,posy);
            XSetForeground(mydisplay,mygc,(pcore->portb&1)?colf:colb);
            posx=x;
            if (debug_status!=RUNNING) XFlush(mydisplay);
        }

        opb=pcore->portb; otrb=pcore->trisb;
}

void do_periodic_update()
{
        joyupdate();
}

void trace_interrupt(core *pcore){}
