/**********************************************************************
Sound.c
This file does not generate an object. It is included in core.c instead.
The reason for this is the sound_tic inline. Inlined functions can't be
located in a diferent object.
***********************************************************************/

#include <fcntl.h>
#include <linux/soundcard.h>

// Sample rates below 12KHz can give amplitude overflows
#define SRATE           22050
#define AUDIO_DEV       "/dev/dsp"
#define SND_PORT        porta
#define SND_BIT         4

// Tris register selection. Do not edit
#if SND_PORT==porta
#define SND_TRIS        trisa
#else
#define SND_TRIS        trisb
#endif

int audio_fd=0;
unsigned int cyclespersample;

void init_sound(int freq)
{
	int fs=0x00030007,rate=SRATE,bits=8,stereo=0;

        cyclespersample=freq/SRATE;

	if ((audio_fd=open(AUDIO_DEV, O_WRONLY)) == -1) {
		fprintf(stderr, "Cannot open sound device");
		exit(0);
	}

        // Use a small fragment to avoid long sound delays

        if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &fs)) {
        	fprintf(stderr, "Cannot set fragment size\n");
        }
        /*--------------------------------------------------------*/
        if (ioctl(audio_fd, SNDCTL_DSP_SAMPLESIZE, &bits) == EOF) {
                fprintf(stderr, "Cannot set %d bits/sample\n",bits);
        }
        if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo) == EOF) {
                fprintf(stderr, "Cannot set %ld channel(s)\n",stereo+1);
        }
        if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &rate) == EOF) {
                fprintf(stderr, "Cannot set rate %d Hz\n",rate);
        }
}

// This function is called on every cycle. It must be inlined for speed.
// The audio output is low-pass filtered to reduce aliasing.
// A very simple comb filter is used for this purpose.

extern inline void sound_tic(core *pcore)
{
        static unsigned int tcount=1,ix=0;
        static unsigned int i1,i2,oi2,do1,do2,odo2;
        static unsigned char buffer[128];

        // Anti-alias filter (2nd order Comb)
        i2+=i1;
        if ((pcore->SND_PORT & (1<<SND_BIT)) &&
            (~(pcore->SND_TRIS & (1<<SND_BIT)))) i1++;

        if(!(--tcount)) {
                tcount=cyclespersample;
                do2=i2-oi2; oi2=i2;
                do1=do2-odo2; odo2=do2;
                do1/=cyclespersample;
                do1<<=7; do1/=cyclespersample;
                buffer[ix++]=do1;
                if (ix==128) {
                        // This write blocks until more space is avaiable
                        // In the output buffer. This gives us a real-time
                        // speed emulation.
                        write(audio_fd,buffer,128);
                        ix=0;
                }
        }
}
